/* main program for DB101 prototype */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <avr/wdt.h>
#include <delay_x_gcc.h>

#include <power_driver.h>
#include <uart_driver.h>
#include <fifo_lib.h>
#include <terminal_lib.h>
#include <backlight_driver.h>

#include "../nmea_reader.h"
#include <sdcmd.h>

#define RXBUFSIZE 100
#define TXBUFSIZE 20

static FIFO_data_t rxbuf[RXBUFSIZE];
static FIFO_data_t txbuf[TXBUFSIZE];

static FIFO_handle_t volatile rxfifo,txfifo;

TERM_terminal_t term;

static void
terminal_init (void)
{

  FIFO_Init (&rxfifo, rxbuf, RXBUFSIZE);
  FIFO_Init (&txfifo, txbuf, TXBUFSIZE);

  LCD_SetScreen (0);
  TERM_Init (&term, 8, 1, true);
  TERM_EnableWriteThrough (&term);
  TERM_Redraw (&term);
}

static void
terminal (void)
{
  UART_Init (&rxfifo, &txfifo);

  do {
    while (!FIFO_HasData (&rxfifo, FIFO_data_t)) {
      POWER_EnterIdleSleepMode ();
    }

    if (FIFO_HasData (&rxfifo, FIFO_data_t)) {
      FIFO_data_t b;
      FIFO_GetData (&rxfifo, &b);
#if DEBUG
      if (!FIFO_HasData (&rxfifo, FIFO_data_t)) {
	_delay_ms(1000.0*40.0*10.0/9600.0);
      }
      CAL_disable_interrupt ();
#endif
      nmea_serial_in_handler (b);
#if DEBUG
      CAL_enable_interrupt ();
#endif

      /*TERM_ProcessChar (&term, b);*/
    }
  } while (1);
  UART_SetRxFifo (NULL);
  UART_SetTxFifo (NULL);
}

static struct speed_table_s {
  uint16_t speed;
  uint16_t ubrrv;
} PROGMEM speed_table [] = {
#define SPEED(s) { s, UART_CALC_UBRR_VALUE(s) }
  SPEED(2400),
  SPEED(4800),
  SPEED(9600),
  SPEED(19200),
#undef SPEED
};

void serial_init (uint16_t speed)
{
  uint16_t ubrrv = 0;
  struct speed_table_s * s;
  uint16_t cspeed;
  for (s = speed_table; (cspeed = pgm_read_word(&s->speed)); s++) {
    if (cspeed == speed) {
      ubrrv = pgm_read_word(&s->ubrrv);
      break;
    }
  }
  if (ubrrv == 0) ubrrv = UART_CALC_UBRR_VALUE(9600);
  PRR1 &= ~(1<<PRUSART1);
  UART_SetBaudRateValue (ubrrv);
  UART_SetParityMode (UART_PARITY_NONE);
  UART_SetDataBits (8);
  UART_SetStopBits (1);
  DDRD |= _BV(PD4); PORTD |= _BV(PD4);
}

void watchdog_off (void) __attribute__((naked)) __attribute__((section(".init3")));
void watchdog_off (void)
{
  wdt_reset ();
  MCUSR &= ~(_BV(WDRF));
  wdt_disable ();
}

static void printterm (const char * s)
{
  TERM_ProcessString (&term, s);
}

uint8_t sdbuf[512][2];

int
main (void)
{
  int i;
#if 0
  InitMemory ();
#endif
  POWER_Init ();
  LCD_Init ();
  BACKLIGHT_Init ();
  BACKLIGHT_SetIntensity (255);
  BACKLIGHT_SetRGB (255, 224, 224);
  nmea_init ();
#if 0
  nmea_start_callback = &printterm;
#endif

  terminal_init ();
  sd_init_card (printterm);
  
  /* wait 20s */
  for (i = 200; i > 0; i--) _delay_ms(100);

  /* read first block from SD */
  if (sd_read_block (sdbuf[0], 0) == 0) {
    /* print first bytes of contents */
    uint8_t i, j;
    uint8_t *p;
    for (p=sdbuf[0], i=0; i<512/8; i++) {
      uint16_t a = i*8;
      TERM_ProcessChar (&term, (a>>8)<10 ? ((a>>8)+'0') : ((a>>8)-10+'A'));
      a &= 0xff;
      TERM_ProcessChar (&term, (a>>4)<10 ? ((a>>4)+'0') : ((a>>4)-10+'A'));
      a &= 0x0f;
      TERM_ProcessChar (&term, a<10 ? (a+'0') : (a-10+'A'));
      TERM_ProcessChar (&term, ' ');
      for (j=0; j<8; j++) {
	uint8_t b = *p++;
	uint8_t c = b >> 4;
	b &= 0xf;
	TERM_ProcessChar (&term, c < 10 ? (c+'0') : (c-10+'A'));
	TERM_ProcessChar (&term, b < 10 ? (b+'0') : (b-10+'A'));
      }
      TERM_ProcessString (&term, "\r\n");
      _delay_ms (200);
    }
  }

  /* wait 20s */
  for (i = 200; i > 0; i--) _delay_ms(100);

  CAL_enable_interrupt ();

  terminal ();

  return 0;
}
