/* event handling */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <avr/interrupt.h>
#include <util/atomic.h>

#include "events.h"
#include "nmea_reader.h"

#define MAX_EVENTS (16)

volatile unsigned char event_queue_cnt /*= 0*/;

/* the queue is used as a ring buffer */
static event queue [MAX_EVENTS];

/* index to first not-handled event */
static unsigned char event_queue_first /*= 0*/;

static void handle_event_p (event *);

void
handle_event (unsigned char n)
{
  unsigned char index;
  if (event_queue_cnt<=n) return;
  
  index = (event_queue_first + n) % MAX_EVENTS;
  handle_event_p (queue+index);
  while (event_queue_cnt && queue[event_queue_first].ev_type == NOP) {
    cli ();
    event_queue_cnt--;
    event_queue_first++;
    if (event_queue_first >= MAX_EVENTS) event_queue_first = 0;
    sei ();
  }
}

static void
handle_event_p (event * e)
{
  switch (e->ev_type) {
  case NOP: break;
  case KEY:
    /* do_key (e->ev_data.k.pressed, e->ev_data.k.keycode); */
    break;
  case TIMER:
    /* do_timer (e->ev_data.t.timer_id); */
    break;
  case SERIAL_IN:
    nmea_serial_in_handler (e->ev_data.s.byte);
    break;
  case SERIAL_OUT:
    /* do_serial_out (); */
    break;
  case NMEA_POS:
    /* do_nmea (lat,lon,height,hdop,vdop) */
    break;
  case SD_READY:
    /* do_sd_card (); */
    break;
  }
}

static uint8_t
prepare_new_event (event_type t)
{
  uint8_t i;
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    if (event_queue_cnt >= MAX_EVENTS)
      return MAX_EVENTS;
    i = event_queue_cnt++;
  }
  i += event_queue_first;
  /*if (i >= MAX_EVENTS)*/
    i %= MAX_EVENTS;
  /* now i is a good index into array */
  queue[i].ev_type = t;
  return i;
}

uint8_t
put_serial_in_event (uint8_t byte)
{
  uint8_t evidx = prepare_new_event (SERIAL_IN);
  if (evidx == MAX_EVENTS) return 0;
  queue[evidx].ev_data.s.byte = byte;
  return 1;
}

uint8_t
put_nmea_event (struct nmea_data_s * p
		/* gps_time_t time, gps_date_t date,
		uint32_t lat, uint16_t latf,
		uint32_t lon, uint16_t lonf, char latsign,
		char lonsign, uint16_t height, uint16_t hdop,
		uint16_t vdop*/)
{
  uint8_t evidx = prepare_new_event (NMEA_POS);
  if (evidx == MAX_EVENTS) return 0;
  queue[evidx].ev_data.p.time = p->time;
  queue[evidx].ev_data.p.date = p->date;
  queue[evidx].ev_data.p.lat = p->latitude >> 16;
  queue[evidx].ev_data.p.latf = p->latitude & -((uint32_t)1<<16);
  queue[evidx].ev_data.p.lon = p->longitude >> 16;
  queue[evidx].ev_data.p.lonf = p->longitude & -((uint32_t)1<<16);
  queue[evidx].ev_data.p.latsign = p->latitude_sign;
  queue[evidx].ev_data.p.lonsign = p->longitude_sign;
  queue[evidx].ev_data.p.height = p->altitude;
  queue[evidx].ev_data.p.hdop = p->hdop;
  queue[evidx].ev_data.p.vdop = p->vdop;
  return 1;
}

uint8_t
put_kbd_i2c_event (uint8_t byte0, uint8_t byte1)
{
  uint8_t evidx = prepare_new_event (KBD_I2C);
  if (evidx == MAX_EVENTS) return 0;
  queue[evidx].ev_data.i2c.byte0 = byte0;
  queue[evidx].ev_data.i2c.byte1 = byte1;
  return 1;
}

uint8_t
put_key_event (uint16_t changed, uint16_t pressed)
{
  uint8_t evidx = prepare_new_event (KEY);
  if (evidx == MAX_EVENTS) return 0;
  queue[evidx].ev_data.k.changed = changed;
  queue[evidx].ev_data.k.pressed = pressed;
  return 1;
}
