/// This file has been prepared for Doxygen automatic documentation generation.
/*! \file *********************************************************************
 *
 * \brief  LCD graphics library source file
 *
 *         This file contains higher-level graphics functions that cannot
 *         directly utilize the page-oriented nature of the LCD screen, and
 *         use the pixel manipulation functions of the lcd_lib.* files instead.
 *         This is an extension to the lcd_lib.* files.
 *
 * \par Application note:
 *      AVR482: DB101 Software
 *
 * \par Documentation
 *      For comprehensive code documentation, supported compilers, compiler
 *      settings and supported devices see readme.html
 *
 * \author
 *      Atmel Corporation: http://www.atmel.com \n
 *      Support email: avr@atmel.com
 *
 * $Revision: 2420 $
 * $URL: http://revisor.norway.atmel.com/AppsAVR8/avr482_db101_software_users_guide/tags/release_A_code_20070917/IAR/gfx/gfx_lib.c $
 * $Date: 2007-09-17 11:57:01 -0600 (ma, 17 sep 2007) $  \n
 *
 * Copyright (c) 2006, Atmel Corporation All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of ATMEL may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY AND
 * SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include "gfx_lib.h"
#include "lcd_lib.h"
#include <stdint.h>


/****************************
 * Function implementations.
 ****************************/

/*
 * \param  Xstart  X-coordinate for the start of the line
 * \param  Ystart  Y-coordinate for the start of the line
 * \param  Xend    X-coordinate for the end of the line
 * \param  Yend    X-coordinate for the end of the line
 */
void LCD_SetLine(uint8_t Xstart, uint8_t Ystart, uint8_t Xend, uint8_t Yend)
{
    uint8_t x, y, i;
    int8_t e, xinc, yinc;
    int8_t dx, dy;

    if (Xstart > Xend){
        dx = Xstart;        /* swap x1,y1  with x2,y2 */
        Xstart = Xend;
        Xend = dx;
        dy = Ystart;
        Ystart = Yend;
        Yend = dy;
    }
	
    dx = Xend - Xstart;
    dy = Yend - Ystart;

    x = Xstart;
    y = Ystart;

    if (dx < 0){
        xinc = -1;
        dx = -dx;
    } else {
	    xinc = 1;
	}
    if (dy < 0){
        yinc = -1;
        dy = -dy;
    } else {
        yinc = 1;
	}

    /****************************************/

    if (dx > dy){
        e = dy - dx;
        for (i = 0; i <= dx; i++)
            {
            /* write out the data */
            LCD_SetPixel(x,y);
            if (e >= 0){
                e -= dx;
                y += yinc;
            } 
            e += dy;
            x += xinc;
        }
    } else {
        e = dx - dy;
        for (i = 0; i <= dy; i++){
            LCD_SetPixel(x,y);
    
            if (e >= 0){
                e -= dy;
                x += xinc;
            }
            e += dx;
            y += yinc;
        }
    }
}


/*
 * \param  Xstart  X-coordinate for the start of the line
 * \param  Ystart  Y-coordinate for the start of the line
 * \param  Xend    X-coordinate for the end of the line
 * \param  Yend    X-coordinate for the end of the line
 */
void LCD_ClrLine(uint8_t Xstart, uint8_t Ystart, uint8_t Xend, uint8_t Yend)
{
    uint8_t x, y, i;
    int8_t e, xinc, yinc;
    int8_t dx, dy;

    if (Xstart > Xend){
        dx = Xstart;        /* swap x1,y1  with x2,y2 */
        Xstart = Xend;
        Xend = dx;
        dy = Ystart;
        Ystart = Yend;
        Yend = dy;
    }

    dx = Xend - Xstart;
    dy = Yend - Ystart;

    x = Xstart;
    y = Ystart;

    if (dx < 0){
        xinc = -1;
        dx = -dx;
    } else {
	    xinc = 1;
	}
    if (dy < 0){
        yinc = -1;
        dy = -dy;
    } else {
        yinc = 1;
	}

    /****************************************/

    if (dx > dy){
        e = dy - dx;
        for (i = 0; i <= dx; i++)
            {
            /* write out the data */
            LCD_ClrPixel(x,y);
            if (e >= 0){
                e -= dx;
                y += yinc;
            } 
            e += dy;
            x += xinc;
        }
    } else {
        e = dx - dy;
        for (i = 0; i <= dy; i++){
            LCD_ClrPixel(x,y);
    
            if (e >= 0){
                e -= dy;
                x += xinc;
            } 
            e += dx;
            y += yinc;
        }
    }
}


/*
 * \param  Xcenter  X-coordinate for the center of the circle
 * \param  Ycenter  Y-coordinate for the center of the circle
 * \param  Radius   Radius of the circle in pixels
 */
void LCD_SetCircle(uint8_t Xcenter, uint8_t Ycenter, uint8_t Radius)
{
    int16_t tswitch;
    uint8_t X, Y;
    uint8_t dYX;

    dYX = Ycenter - Xcenter;
    X = 0;
    Y = Radius;
    tswitch = 3 - 2 * Radius;
    while (X <= Y) {
        LCD_SetPixel(Xcenter + X, Ycenter + Y);     
        LCD_SetPixel(Xcenter + X, Ycenter - Y);
        LCD_SetPixel(Xcenter - X, Ycenter + Y);     
        LCD_SetPixel(Xcenter - X, Ycenter - Y);
        LCD_SetPixel(Ycenter + Y - dYX, Ycenter + X); 
        LCD_SetPixel(Ycenter + Y - dYX, Ycenter - X);
        LCD_SetPixel(Ycenter - Y - dYX, Ycenter + X); 
        LCD_SetPixel(Ycenter - Y - dYX, Ycenter - X);

        if (tswitch < 0) {
            tswitch += (4 * X + 6);
        }
        else {
            tswitch += (4 * (X - Y) + 10);
            Y--;
        }
        X++;
    }
}


/*
 * \param  Xcenter  X-coordinate for the center of the circle
 * \param  Ycenter  Y-coordinate for the center of the circle
 * \param  Radius   Radius of the circle in pixels
 */
void LCD_ClrCircle(uint8_t Xcenter, uint8_t Ycenter, uint8_t Radius)
{
    int16_t tswitch;
    uint8_t X, Y;
    uint8_t dYX;

    dYX = Ycenter - Xcenter;
    X = 0;
     Y = Radius;
    tswitch = 3 - 2 * Radius;
    while (X <= Y) {
        LCD_ClrPixel(Xcenter + X, Ycenter + Y);     
        LCD_ClrPixel(Xcenter + X, Ycenter - Y);
        LCD_ClrPixel(Xcenter - X, Ycenter + Y);     
        LCD_ClrPixel(Xcenter - X, Ycenter - Y);
        LCD_ClrPixel(Ycenter + Y - dYX, Ycenter + X); 
        LCD_ClrPixel(Ycenter + Y - dYX, Ycenter - X);
        LCD_ClrPixel(Ycenter - Y - dYX, Ycenter + X); 
        LCD_ClrPixel(Ycenter - Y - dYX, Ycenter - X);

        if (tswitch < 0) {
            tswitch += (4 * X + 6);
        }
        else {
            tswitch += (4 * (X - Y) + 10);
            Y--;
        }
        X++;
    }
}
