// This file has been prepared for Doxygen automatic documentation generation.
/*! \file *********************************************************************
 *
 * \brief  Single UART driver source file
 *
 *      The UART driver provides an interface to one of the UART peripherals.
 *      The default implementation for the DB101 firmware is to interface the
 *      UART1 module, which is connected to the external UART connections of the
 *      DB101 board.
 *
 *      The driver provides the basic functionality for interrupt controlled
 *      reception and transmission through FIFO queues from the FIFO Queue
 *      Library, and of course to configure the UART baud rate, parity mode etc.
 *
 *      The driver source code is designed to support only one UART module,
 *      which means that it cannot change UART modules runtime. This limitation
 *      makes the driver very speed efficient, but less flexible. For the DB101
 *      board, this is not an issue, since the other UART module is solely used
 *      for accessing the on-board DataFlash memory device.
 *
 * \par Application note:
 *      AVR482: DB101 Software
 *
 * \par Documentation
 *      For comprehensive code documentation, supported compilers, compiler
 *      settings and supported devices see readme.html
 *
 * \author
 *      Atmel Corporation: http://www.atmel.com \n
 *      Support email: avr@atmel.com
 *
 * $Revision: 2436 $
 * $URL: http://revisor.norway.atmel.com/AppsAVR8/avr482_db101_software_users_guide/tags/release_A_code_20070917/IAR/uart_driver/uart_driver.c $
 * $Date: 2007-09-17 15:18:48 -0600 (ma, 17 sep 2007) $  \n
 *
 * Copyright (c) 2006, Atmel Corporation All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. The name of ATMEL may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY AND
 * SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include <stdint.h>

#include <cal.h>
#include <common.h>

#include "uart_driver.h"
#include "fifo_lib.h"


/*******************
 * Private defines.
 *******************/


#if defined(UART)
	#define UART_DATA    UDR
	#define UART_CTRLA   UCSRA
	#define UART_CTRLB   UCSRB
	#define UART_CTRLC   UCSRC
	#define UART_UBRRL	UBRRL
	#define UART_UBRRH	UBRRH
	#define UART_RX_vector USART_RXC_vect
	#define UART_TX_vector USART_TXC_vect
	#define UART_UDRE_vector USART_UDRE_vect
#elif defined(UART0)
	#define UART_DATA    UDR0
	#define UART_CTRLA   UCSR0A
	#define UART_CTRLB   UCSR0B
	#define UART_CTRLC   UCSR0C
	#define UART_UBRR    UBRR0
	#if defined(USART0_RX_vect)
		#define UART_RX_vector USART0_RX_vect
		#define UART_TX_vector USART0_TX_vect
		#define UART_UDRE_vector USART0_UDRE_vect
	#elif defined(USART_RX_vect)
		#define UART_RX_vector USART_RX_vect
		#define UART_TX_vector USART_TX_vect
		#define UART_UDRE_vector USART_UDRE_vect
	#endif
#elif defined(UART1)
	#define UART_DATA    UDR1
	#define UART_CTRLA   UCSR1A
	#define UART_CTRLB   UCSR1B
	#define UART_CTRLC   UCSR1C
	#define UART_UBRR    UBRR1
	#define UART_RX_vector USART1_RX_vect
	#define UART_TX_vector USART1_TX_vect
	#define UART_UDRE_vector USART1_UDRE_vect
#elif defined(UART2)
	#define UART_DATA    UDR2
	#define UART_CTRLA   UCSR2A
	#define UART_CTRLB   UCSR2B
	#define UART_CTRLC   UCSR2C
	#define UART_UBRR    UBRR2
	#define UART_RX_vector USART2_RX_vect
	#define UART_TX_vector USART2_TX_vect
	#define UART_UDRE_vector USART2_UDRE_vect
#elif defined(UART3)
	#define UART_DATA    UDR3
	#define UART_CTRLA   UCSR3A
	#define UART_CTRLB   UCSR3B
	#define UART_CTRLC   UCSR3C
	#define UART_UBRR    UBRR3
	#define UART_RX_vector USART3_RX_vect
	#define UART_TX_vector USART3_TX_vect
	#define UART_UDRE_vector USART3_UDRE_vect
#else
	#error "Please define UART, UART0, UART1, UART2, or UART3"
#endif

/* Since the IAR compiler can handle access to 16-bit registers, as long as they 
 are placed next to each other, UBRR can be used instead of UBRRL and UBRRH */
#ifdef UART
	#define UART_SetUBRR(value)\
		UART_UBRRL = (uint8_t)(value);\
		UART_UBRRH = (uint8_t)(value>>8)
#else
	#define UART_SetUBRR(value)\
		UART_UBRR = value
#endif		


/* Defines to ensure that bit names are defined - substitusions can be allowed
   since bits are located in the same addresses for all UARTs. If one is undefined, all are. */
#ifndef RXC
	#define RXC RXC0
	#define TXC TXC0
	#define UDRE UDRE0
	#define FE FE0
	#define DOR DOR0
	#define UPE UPE0
	#define U2X U2X0
	#define MPCM MPCM0
	#define RXCIE RXCIE0
	#define TXCIE TXCIE0
	#define UDRIE UDRIE0
	#define RXEN RXEN0
	#define TXEN TXEN0
	#define UCSZ2 UCSZ02
	#define RXB8 RXB80
	#define TXB8 TXB80
	#define UMSEL1 UMSEL01
	#define UMSEL0 UMSEL00
	#define UPM1 UPM01
	#define UPM0 UPM00
	#define USBS USBS0
	#define UCSZ1 UCSZ01
	#define UCSZ0 UCSZ00
	#define UCPOL UCPOL0
#endif


/*********************
 * Private variables.
 *********************/


FIFO_handle_t * rxFifo; //!< Fifo buffers used when receiving data.
FIFO_handle_t * txFifo;	//!< Fifo buffers used when transmitting data.


//! Control and status flags for UART (access though macros)
UART_flags_t volatile UART_flags;


/****************************
 * Function implementations.
 ****************************/

/*! \brief RX complete interrupt handler.
 * 
 * Puts received data in RX Fifo. If Fifo is full the byte is lost, and the 
 * overflow flag is set (use UART_GetOverRxOverflowFlag to test overflow flag).
 *
 */
CAL_ISR( UART_RX_vector )
{
	uint8_t data = UART_DATA;
	if (FIFO_HasRoomFor( rxFifo, data )) {
		FIFO_PutUint8( rxFifo, data );
	} else {
		UART_SetRxOverflowFlag();
	}
}



/*! \brief UART data register empty interrupt handler.
 *
 * Reads from TX fifo until the fifo is empty. If fifo is empty the UDRE interrupt
 * is disabled. Call UART_StartTransmission() macro to start a new transmission.
 *
 */
CAL_ISR( UART_UDRE_vector )
{
	if( (txFifo != NULL) && FIFO_HasData(txFifo, char) )
	{
		UART_DATA = FIFO_GetUint8( txFifo );
	}
	else
	{
		UART_DisableUdreInterrupt();
	}
}



/*! \brief TX complete interrupt handler - not used.
 *
 */
CAL_ISR( UART_TX_vector )
{
	CAL_no_operation();
}



/*!
 * The U(S)ART is enabled in standard 1-8-1 mode. Also enables RX interrupt.
 * UART_StartTransmission() can be used to enable TX interrupt.
 *
 *  \param  *rxFifoHandle  Pointer to FIFO where data should be saved
 *  \param  *txFifoHandle  Pointer to FIFO where data should be sent from
 */
void UART_Init( FIFO_handle_t * rxFifoHandle, FIFO_handle_t * txFifoHandle )
{
	rxFifo = rxFifoHandle;
	txFifo = txFifoHandle;
	
	FIFO_Flush( rxFifo );
	FIFO_Flush( txFifo );
	
	UART_CTRLB |= (1<<RXEN)|(1<<TXEN);	
	UART_EnableRxInterrupt();
}


/*!
 * \param  ubrrValue  New baudrate value
 */
void UART_SetBaudRateValue( uint16_t ubrrValue )
{
	UART_SetUBRR( ubrrValue );
}


/*! 
 * \param  parityMode  A UART_parityMode_t enum
*/
void UART_SetParityMode( UART_parityMode_t parityMode )
{
	switch (parityMode) {
		case UART_PARITY_NONE :
			UART_CTRLC = ( UART_CTRLC &
			             ~((1 << UPM1) | (1 << UPM0)) ) |
			             ((0 << UPM1) | (0 << UPM0));
		break;

		case UART_PARITY_EVEN :
			UART_CTRLC = ( UART_CTRLC &
			             ~((1 << UPM1) | (1 << UPM0)) ) |
			             ((1 << UPM1) | (0 << UPM0));
		break;

		case UART_PARITY_ODD :
			UART_CTRLC = ( UART_CTRLC &
			             ~((1 << UPM1) | (1 << UPM0)) ) |
			             ((1 << UPM1) | (1 << UPM0));
		break;
		
		default: break;
	}
}

/*!
 * Does not change anything if number not is supported
 *
 * \param  dataBitCount  Number of data bits
*/
void UART_SetDataBits( uint8_t dataBitCount )
{
	switch (dataBitCount) {
		case 5 :
			UART_CTRLB &= ~(1 << UCSZ2);
			UART_CTRLC = ( UART_CTRLC &
			             ~((1 << UCSZ1) | (1 << UCSZ0)) ) |
			             ((0 << UCSZ1) | (0 << UCSZ0));
		break;

		case 6 :
			UART_CTRLB &= ~(1 << UCSZ2);
			UART_CTRLC = ( UART_CTRLC &
			             ~((1 << UCSZ1) | (1 << UCSZ0)) ) |
			             ((0 << UCSZ1) | (1 << UCSZ0));
		break;

		case 7 :
			UART_CTRLB &= ~(1 << UCSZ2);
			UART_CTRLC = ( UART_CTRLC &
			             ~((1 << UCSZ1) | (1 << UCSZ0)) ) |
			             ((1 << UCSZ1) | (0 << UCSZ0));
		break;

		case 8 :
			UART_CTRLB &= ~(1 << UCSZ2);
			UART_CTRLC = ( UART_CTRLC &
			             ~((1 << UCSZ1) | (1 << UCSZ0)) ) |
			             ((1 << UCSZ1) | (1 << UCSZ0));
		break;

		case 9 :
			UART_CTRLB |= (1 << UCSZ2);
			UART_CTRLC = ( UART_CTRLC &
			             ~((1 << UCSZ1) | (1 << UCSZ0)) ) |
			             ((1 << UCSZ1) | (1 << UCSZ0));
		break;

		default : break;
	}
}


/*!
 * Does not change anything if number not is supported
 *
 * \param stopBitCount  Number of stop bits
*/
void UART_SetStopBits( uint8_t stopBitCount )
{
	switch (stopBitCount) {
		case 1 : UART_CTRLC &= ~(1 << USBS); break;
		case 2 : UART_CTRLC |= (1 << USBS); break;
		default : break;
	}
}


/*! 
 *  \param  *rxFifoHandle  Is a pointer to the RX Fifo handle
 */
void UART_SetRxFifo(FIFO_handle_t *rxFifoHandle)
{
	rxFifo = rxFifoHandle;
}


/*!
 *  \param  *txFifoHandle  Is a pointer to the TX Fifo handle
 */
void UART_SetTxFifo(FIFO_handle_t *txFifoHandle)
{
	txFifo = txFifoHandle;
}


/*!
 *  \return  A pointer to the RX Fifo handle
 */
FIFO_handle_t* UART_GetRxFifo( void )
{
	return rxFifo;
}


/*!
 *  \return  Is a pointer to the TX Fifo handle
 */
FIFO_handle_t* UART_GetTxFifo( void )
{
	return txFifo;
}
