/* SD card access functions */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <sdcmd.h>
#include <spi.h>
#include <util/crc16.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h> /* for sprintf() */
#include "crc7.h"

#define spi_sdread() spi_read(0xff)

static uint8_t sd_select (void)
{
  uint8_t tok = 0xff;
  uint8_t timeout;
  spi_select ();
#if 0
  for (timeout=1; timeout && tok == 0xff; timeout--) {
    tok = spi_sdread ();
  }
#endif
  return tok;
}

static void sd_deselect (void)
{
  spi_write (0xff);
  spi_deselect ();
  spi_write (0xff); /* some more clocks for card to really deselect */
}

/* for debugging */
static void(*print)(const char *);
#define BUFSIZE (10+1)
static char buf[BUFSIZE];

static void debug_msg(const char * msg)
{
  if (print) {
    (*print)(msg);
  }
}

static char int2hex (uint8_t i)
{
  if (i < 10) return '0'+i;
  return 'A'+i-10;
}

static void debug_byte(uint8_t b)
{
  if (print) {
    buf[0] = int2hex(b>>4);
    buf[1] = int2hex(b&15);
    buf[2] = '\0';
    debug_msg(buf);
  }
}

/* CRC: CCITT-16 */
/* compute CRC for 136bit response block */
static uint16_t sd_resp136_crc (uint8_t sd_response[])
{
  uint8_t i;
  uint16_t crc = 0x0000;
  for (i=0; i<=16; i++)
    crc = _crc_ccitt_update (crc, sd_response[i]);
  return crc;
}

/* for debugging: */
#if 0
#undef spi_write
#define spi_write(b) do {debug_byte(spi_transfer((b),1,1));} while (0)
#endif

static void sd_send_cmd (uint8_t cmd, uint32_t arg)
{
  uint8_t crc = 0;
  uint8_t a1 = (arg>>24), a2 = (arg>>16), a3 = (arg>>8), a4 = (arg);

#if 0 /* expect cmd in finished format */
  cmd &= 0x3f; cmd |= 0x40; /* clear start bit, set cmd bit */
#endif

  spi_write (cmd); crc = _crc7_update (crc, cmd);
  spi_write (a1); crc = _crc7_update (crc, a1);
  spi_write (a2); crc = _crc7_update (crc, a2);
  spi_write (a3); crc = _crc7_update (crc, a3);
  spi_write (a4); crc = _crc7_update (crc, a4);
  spi_write ((crc<<1)|1);

  if (print) {
    debug_msg("SD");
    debug_byte(cmd);
    debug_msg(" ");
    debug_byte(a1); debug_byte(a2); debug_byte(a3);debug_byte(a4);
  }
}

static void sd_busy_wait (void)
{
  while (spi_sdread () == 0) /* wait until !busy */;
}

static uint8_t sd_recv_resp (void)
{
  uint8_t r;
  uint8_t timeout = 10;
  do {
    r = spi_sdread ();
  } while (r == 0xff && timeout-- > 0);
  return r;
}

static uint8_t sd_recv_response1 (void)
{
  uint8_t r = sd_recv_resp ();
  if (print) {
    debug_msg("R");
    debug_byte(r);
    debug_msg("\r\n");
  }
  return r;
}

static uint8_t sd_recv_response1b (void)
{
  uint8_t r = sd_recv_response1 ();
  sd_busy_wait ();
  return r;
}

static uint16_t sd_recv_response2 (void)
{
  uint16_t a = sd_recv_resp ();
  if (a != 0xff)
    return (a << 8) | spi_sdread ();
  else
    return 0xffff;
}

static uint8_t sd_recv_response3 (uint32_t * pocr)
{
  uint8_t r1 = sd_recv_response1 ();
  if (r1 != 0xff) {
    uint32_t ocr;
    ocr = (uint32_t)spi_sdread () << 24;
    ocr |= (uint32_t)spi_sdread () << 16;
    ocr |= (uint16_t)spi_sdread () << 8;
    ocr |= spi_sdread ();
    *pocr = ocr;
  }
  return r1;
}

static uint8_t sd_recv_response7 (uint32_t * result)
{
  /* same as R3 */
  return sd_recv_response3 (result);
}

static uint8_t sd_recv_data_response (void)
{
  return sd_recv_resp ();
}

static uint8_t sd_send_acmd (uint8_t acmd, uint32_t arg)
{
  uint8_t r;
  sd_send_cmd (SD_APP_CMD, 0);
  r = sd_recv_response1 ();
  sd_deselect (); /* deselect/select between CMDs */
  sd_select ();
  if (r & (1<<2)) return r;
  sd_send_cmd (acmd, arg);
  return r;
}

static uint8_t sd_recv_data (uint8_t *data, sd_blen_t blocklen)
{
  uint16_t crc = 0;
  uint16_t crc_read;
  uint8_t t;

  do {
    t = spi_transfer (0xff, 0, 1);
  } while (t == 0xff); /* wait for start block token */

  for (; blocklen; blocklen--) {
    uint8_t b = spi_transfer (0xff, 0, 1);
    *data++ = b;
    crc = _crc16_update (crc, b);
  }
  crc_read = spi_transfer (0xff, 0, 1); crc_read <<= 8;
  crc_read |= spi_transfer (0xff, 0, 1);
  return crc_read == crc;
}

static void sd_send_data (uint8_t *data, sd_blen_t blocklen, uint8_t token)
{
  uint16_t crc = 0;
  spi_write (token);
  for (; blocklen; blocklen--) {
    uint8_t b = *data++;
    spi_write (b);
    crc = _crc16_update (crc, b);
  }
  if (data) {
    /* only send CRC when we actually did send some data */
    spi_write (crc >> 8);
    spi_write (crc & 0xff);
  }
}

uint8_t sd_go_idle (void)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_GO_IDLE_STATE, 0);
  r = sd_recv_response1 ();
  sd_deselect ();
  return r;
}

uint8_t sd_get_if_cond (uint8_t *version, uint8_t *vhs, uint8_t * check)
{
  uint8_t r;
  uint32_t tmp = ((uint16_t)(*vhs & 0x0f) << 8) | *check;
  sd_select ();
  sd_send_cmd (SD_SEND_IF_COND, tmp);
  r = sd_recv_response7 (&tmp);
  sd_deselect ();
  *version = (tmp >> 28) & 0x0f;
  *vhs = (tmp >> 8) & 0x0f;
  *check = (tmp) & 0xff;
  return r;
}

static uint8_t sd_get_csd (sd_csd_data_t * csd)
{
  uint8_t r;

  sd_select ();
  sd_send_cmd (SD_SEND_CSD, 0);
  r = sd_recv_response1 ();
  if (r == 0) {
    sd_recv_data ((uint8_t*)csd, sizeof (*csd));
  }
  sd_deselect ();

  return r;
}

uint8_t sd_get_cid (sd_cid_data_t * cid)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_SEND_CID, 0);
  r = sd_recv_response1 ();
  if (r == 0) {
    sd_recv_data ((uint8_t *)cid, sizeof (*cid));
  }
  sd_deselect ();
  return r;
}

static uint8_t sd_stop_trans (void)
{
  uint8_t r;
  /* we already _are_ in a transaction */
  sd_send_cmd (SD_STOP_TRANSMISSION, 0);
  r = sd_recv_response1b ();
  return r;
}

uint16_t sd_get_status (void)
{
  uint16_t r;
  sd_select ();
  sd_send_cmd (SD_SEND_STATUS, 0);
  r = sd_recv_response2 ();
  sd_deselect ();
  return r;
}

sd_blen_t sd_current_blocklen = 512;

uint8_t sd_set_blocklen (sd_blen_t blocklen)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_SET_BLOCKLEN, blocklen);
  r = sd_recv_response1 ();
  sd_deselect ();
  if (r == 0) sd_current_blocklen = blocklen;
  return r;
}

uint8_t sd_read_block (uint8_t *data, uint32_t address)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_READ_SINGLE_BLOCK, address);
  r = sd_recv_response1 ();
  if (r == 0) {
    sd_recv_data (data, sd_current_blocklen);
  }
  sd_deselect ();
  return r;
}

uint8_t sd_read_multiple_block (uint8_t *data[], uint8_t cnt, uint32_t address)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_READ_MULTIPLE_BLOCK, address);
  r = sd_recv_response1 ();
  if (r == 0) {
    for (; cnt; cnt--) {
      r = sd_recv_data (*data++, sd_current_blocklen);
      if ((r & 0b00011111) != 0b00000101) break;
    }
    r = sd_stop_trans ();
  }
  sd_deselect ();
  return r;
}

uint8_t sd_write_block (uint8_t *data, uint32_t address)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_WRITE_BLOCK, address);
  r = sd_recv_response1 ();
  if (r == 0) {
    sd_send_data (data, sd_current_blocklen, SDTOKEN_START);
    r = sd_recv_data_response ();
  }
  sd_deselect ();
  return r;
}

uint8_t sd_write_multiple_block (uint8_t *data[], uint8_t cnt, uint32_t address)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_READ_MULTIPLE_BLOCK, address);
  r = sd_recv_response1 ();
  if (r == 0) {
    for (; cnt; cnt--, data) {
      sd_send_data (*data++, sd_current_blocklen, SDTOKEN_START_MULTIPLE);
      r = sd_recv_data_response ();
      if (r) break;
    }
    sd_send_data (0, 0, SDTOKEN_STOP_MULTIPLE);
    sd_busy_wait ();
  }
  sd_deselect ();
  return r;
}

uint8_t sd_program_csd (sd_csd_data_t *csd)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_PROGRAM_CSD, 0);
  r = sd_recv_response1 ();
  if (r == 0) {
    sd_send_data ((uint8_t *)csd, sizeof (*csd), SDTOKEN_START);
    r = sd_recv_data_response ();
  }
  sd_deselect ();
  return r;
}

uint8_t sd_get_ocr (uint32_t *ocr)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_READ_OCR, 0);
  r = sd_recv_response3 (ocr);
  sd_deselect ();
  return r;
}

static uint8_t sd_crc_on_off (uint8_t crc_enable)
{
  uint8_t r;
  sd_select ();
  sd_send_cmd (SD_CRC_ON_OFF, crc_enable ? (1<<0) : 0);
  r = sd_recv_response1 ();
  sd_deselect ();
  return r;
}

static uint8_t sd_set_op_cond (uint8_t high_capacity)
{
  uint8_t r;
  sd_select ();
  r = sd_send_acmd (SD_A_SEND_OP_COND, high_capacity ? ((uint32_t)1<<30) : 0);
  if ((r & (1<<2)) == 0) {
    r = sd_recv_response1 ();
  }
  sd_deselect ();
  return r;
}

/* round up */
#define T(v) (VAL_CSD_TRAN_SPEED(v) ?					\
	      (uint8_t)(((float)F_CPU+VAL_CSD_TRAN_SPEED(v)-1)/VAL_CSD_TRAN_SPEED(v)) : \
	      0xff)
#define T1(v) T(v|0), T(v|1), T(v|2), T(v|3), T(v|4), T(v|5), T(v|6), T(v|7)
static prog_uint8_t sd_tran_speed_to_spi_div[128] = {
  T1(0x00), T1(0x08),
  T1(0x10), T1(0x18),
  T1(0x20), T1(0x28),
  T1(0x30), T1(0x38),
  T1(0x40), T1(0x48),
  T1(0x50), T1(0x58),
  T1(0x60), T1(0x68),
  T1(0x70), T1(0x78),
};
#undef T1
#undef T

static uint8_t get_div_for_speed(uint8_t tran_speed)
{
  return pgm_read_byte (&sd_tran_speed_to_spi_div[tran_speed & 0x7f]);
}

struct {
  uint8_t version;
  uint8_t vhs;
  uint8_t check;
  uint8_t proto_version:2;
  uint8_t proto_hc:1;
} sd_if_cond;

uint32_t sd_ocr;
sd_cid_data_t sd_cid;
sd_csd_data_t sd_csd;

uint16_t sd_init_card (void(*p)(const char *))
{
  uint8_t i;
  uint8_t r;
  uint8_t chk;
  void (*save_p)(const char *) = print;
  print = p;
  spi_init (get_div_for_speed((9<<3)|0)); /* SPIfreq <= 400kHz */
  /* SD spec says to wait at least 74 clock cycles after power-up */
  debug_msg("SD_INIT");
  for (i=0; i<10; i++) spi_write (0xff);
  debug_msg(".");
  spi_select (); /* no wait for ready */
  for (i=0; i<10; i++) spi_write (0xff);
  sd_deselect ();
  debug_msg(".\r\n");
  i = 0;
  do {
    _delay_ms(5);
    debug_byte (i++);
    r = sd_go_idle ();
  } while (r & 0b11110000);
  do {
    /* the card needs some time after power-up */
    sd_if_cond.vhs = 0b0001;
    sd_if_cond.check = chk = 0b10101010;
    r = sd_get_if_cond (&sd_if_cond.version, &sd_if_cond.vhs, &sd_if_cond.check);
  } while (r & 0b00000010);
  if (r & 0b00000100) {
    /* SEND_IF_COND not supported, version 1.x */
    sd_if_cond.proto_version = 1;
  } else {
    sd_if_cond.proto_version = 2;
    debug_byte (sd_if_cond.version);
    debug_byte (sd_if_cond.vhs);
    debug_byte (sd_if_cond.check);
    debug_msg ("\r\n");
  }

  /* read Operation Condition Register */
  r = sd_get_ocr (&sd_ocr);
  debug_byte ((sd_ocr & 0xFF0000) >> 16);
  debug_byte (sd_ocr >> 30);
  debug_msg ("\r\n");

  /* initiate startup, set HC mode if v2.x */
  do {
    r = sd_set_op_cond (sd_if_cond.proto_version >= 2);
  } while ((r & (1<<0)) && (r & (1<<2)) == 0);

  /* enable CRC checking, disabled by default in SPI mode */
  r = sd_crc_on_off (1);

  r = sd_get_csd (&sd_csd);
  
  if (sd_if_cond.proto_version >= 2) {
    r = sd_get_ocr (&sd_ocr);
    debug_byte ((sd_ocr & 0xFF0000) >> 16);
    debug_byte (sd_ocr >> 30);
    debug_msg ("\r\n");

    debug_msg ("SD2.0 ");
    if (sd_ocr & ((uint32_t)1<<30)) {
      sd_if_cond.proto_hc = 1;
      debug_msg ("HC");
    } else {
      sd_if_cond.proto_hc = 0;
      debug_msg ("SC");
    }
  } else {
    debug_msg ("SD1.x");
  }

  if (CSD_STRUCTURE(sd_csd) == 0 || CSD_STRUCTURE(sd_csd) == 1) {
    uint32_t capacity = CSD_CARD_CAPACITY_KB(sd_csd);
    uint8_t spi_div;
    char buf[16];
    /* convert and round to MB */
    capacity >>= 9;
    capacity += capacity & 1;
    capacity >>= 1;

    sprintf_P(buf, PSTR(" %5ld MB"), capacity);
    debug_msg (buf);

    /* switch to max. allowed SPI speed */
    spi_div = get_div_for_speed(sd_csd.tran_speed);

#if 1
#define MIN_DIV 4
    spi_write (0xff);
    spi_set_speed (spi_div >= MIN_DIV ? spi_div : MIN_DIV);
    spi_write (0xff);
    spi_select ();
    spi_write (0xff);
    spi_write (0xff);
    spi_deselect ();
#endif
  }
  debug_msg ("\r\n");

  r = sd_get_cid (&sd_cid);
  if (r == 0) {
    char buf[64];
    sprintf_P (buf, PSTR("%02x %.2s %.5s %c.%c %8lx %c/%02d\r\n"),
	       sd_cid.mid, sd_cid.oid, sd_cid.pnm,
	       int2hex (sd_cid.prv >> 4), int2hex (sd_cid.prv & 0xf),
	       sd_cid.psn,
	       int2hex (sd_cid.mdt[1] & 0xf), ((sd_cid.mdt[1] >> 4)|(sd_cid.mdt[0] << 4)) & 0xff);
    debug_msg (buf);
  }

  print = save_p;
  return r;
}
