/* SPI interface header file */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#ifndef SPI_H
#define SPI_H

#include <stdint.h>
#include <avr/io.h>
#include <util/delay.h>

/* port and pin definitions */
#if defined(__AVR_ATmega1281__) || defined(__AVR_ATmega1280__)
#define DDR_SPI    DDRB
#define DR_SPI_SS    DDB0
#define DR_SPI_SCK   DDB1
#define DR_SPI_MOSI  DDB2
#define DR_SPI_MISO  DDB3

#define PORT_SPI   PORTB
#define PORT_SPI_SS   PB0
#define PORT_SPI_SCK  PB1
#define PORT_SPI_MOSI PB2
#define PORT_SPI_MISO PB3
#else
#    warning "device type not defined"
#endif


/* initialize SPI interface */
void spi_init (uint8_t div);
/* set SPI speed divider */
void spi_set_speed (uint8_t div);
/* div: divider for SPI clock
 *      possible values: 2 4 8 16 32 64 128
 *         gets rounded up to next possible value
 */

/* transfer one byte via SPI
 * b     byte to write
 * wait  true when we are not sure the previous transfer has already completed
 * read  caller is interested in read byte
 *
 * it is expected that wait and read are constants so the compiler can
 * optimize out the loop and the register read when not needed
 */
static inline uint8_t spi_transfer (uint8_t b, uint8_t wait, uint8_t read)
{
  uint8_t spsr;

  SPDR = b;
  if (wait && (spsr = SPSR) & _BV(WCOL)) {
    /* after write collision just wait for SPIF */
    while (!(spsr & _BV(SPIF))) spsr = SPSR;
    SPDR = b;
  }
  /* at this point, we started the transfer */
  if (read) {
    /* wait for completion */
    while (!(spsr & _BV(SPIF))) spsr = SPSR;
    return SPDR;
  } else {
    /* we don't care when the transfer finishes */
    return 0xff;
  }
}

#define spi_write(b) spi_transfer ((b), 1, 0)
#define spi_read(b)   spi_transfer (b, 1, 1)

/* control of SS
 * ss_enable  true when SS should be active (=low)
 */
static inline void spi_do_select (uint8_t ss_enable)
{
  if (!ss_enable) /* when selecting, we expect the previous transmission to
		   * be complete (we already waited at de-select time) */
    while (0 && !(SPSR & _BV(SPIF)));
  if (ss_enable) {
    PORT_SPI &= ~_BV(PORT_SPI_SS);
  }
  else
    PORT_SPI |= _BV(PORT_SPI_SS);
}

#define spi_select() spi_do_select (1)
#define spi_deselect() spi_do_select (0)

#endif
