/* nmea parser - push style */
/* Copyright (C) 2009 Holger Dietze
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License along
 *   with this program; if not, write to the Free Software Foundation, Inc.,
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <avr/pgmspace.h>
#include <string.h>
#include "nmea_reader.h"
#include "serial.h"
#include "events.h"

void (*nmea_start_callback) (const char *);

uint8_t
nmea_init (void)
{
  serial_init (NMEA_SPEED);
  return 0;
}

typedef enum {
  NMP_TYPE,
  NMP_FIELDS,
  NMP_CSUM,
  NMP_LINEEND,
} nmea_fieldidx_t;

typedef enum {
  NMS_NEW = 0, /* we have _never_ seen any valid start byte */
  NMS_CR_SEEN,
  NMS_LF_SEEN,
  NMS_STARTED,

  NMS_CSUM1,
  NMS_CSUM2,
  NMS_CSUM_SEEN,

  NMS_ERROR, /* parse error encountered */
} nmea_status_t;

typedef enum {
  NMN_NONUMBER = 0,
  NMN_NUMBER_INT,
  NMN_NUMBER_FRAC,
} nmea_number_status_t;

static nmea_status_t nmea_status;

typedef enum {
  NMF_END,
  NMF_IGNORE,
  NMF_TIME,
  NMF_UTCOFFSET_H, NMF_UTCOFFSET_MIN,
  NMF_DATE,
  NMF_DAY, NMF_MONTH, NMF_YEAR,
  NMF_LATITUDE, NMF_LATITUDE_SIGN,
  NMF_LONGITUDE, NMF_LONGITUDE_SIGN,
  NMF_POSITIONFIX,
  NMF_NUM_SATELLITES,
  NMF_PDOP,
  NMF_HDOP, NMF_VDOP,
  NMF_MODE1, NMF_MODE2,
  NMF_SATUSED0, NMF_SATUSED1, NMF_SATUSED2, NMF_SATUSED3,
  NMF_SATUSED4, NMF_SATUSED5, NMF_SATUSED6, NMF_SATUSED7,
  NMF_SATUSED8, NMF_SATUSED9, NMF_SATUSED10, NMF_SATUSED11,
  NMF_GPS_STATUS, NMF_GPS_MODE,
  NMF_ALTITUDE_MSL, NMF_ALTITUDE_UNIT,
  NMF_GEOID_SEPARATION, NMF_GEOID_SEP_UNIT,
  NMF_DGPS_AGE, NMF_DGPS_REFID,
  NMF_GSV_NUMMSG, NMF_GSV_MSG,
  NMF_ELEVATION0, NMF_AZIMUTH0, NMF_SNR0,
  NMF_ELEVATION1, NMF_AZIMUTH1, NMF_SNR1,
  NMF_ELEVATION2, NMF_AZIMUTH2, NMF_SNR2,
  NMF_ELEVATION3, NMF_AZIMUTH3, NMF_SNR3,
  NMF_SPEED, NMF_SPEED_UNIT, NMF_SPEED2, NMF_SPEED2_UNIT,
  NMF_COURSE, NMF_COURSE_BASE, NMF_COURSE2, NMF_COURSE2_BASE,
  NMF_MAGNETIC_VARIATION, NMF_MAGNETIC_VARIATION_SIGN,
  NMF_OK_TO_SEND,
} nmea_field_t;

typedef enum {
  NMO_IGNORE,
  NMO_CHAR,
  NMO_DIGIT, NMO_2DIGIT, NMO_4DIGIT,
  NMO_INT,
  NMO_NUMBER,
  NMO_TIME,
  NMO_DATE_YY,
  NMO_DEG,
  NMO_NORTHSOUTH, /* 'N'->(+) 'S'->(-) */
  NMO_EASTWEST,   /* 'E'->(+) 'W'->(-) */
  NMO_UNIT,       /* 'M'->metric */
} nmea_format_t;

typedef struct {
  nmea_field_t field;
  nmea_format_t format;
} PROGMEM nmea_field_description_t;

/* we rely on that function to be _not_ defined */
extern void _fieldsize_not_defined_ (void);

static inline nmea_field_t pgm_read_field (nmea_field_description_t * d)
{
  switch (sizeof(d->field)) {
  case 1: return pgm_read_byte (&d->field);
  case 2: return pgm_read_word (&d->field);
  default: _fieldsize_not_defined_ ();
  }
}

static inline nmea_format_t pgm_read_format (nmea_field_description_t * d)
{
  switch (sizeof(d->format)) {
  case 1: return pgm_read_byte (&d->format);
  case 2: return pgm_read_word (&d->format);
  default: _fieldsize_not_defined_ ();
  }
}

#define DESC(name) static const char name##_str[] PROGMEM = #name; \
                   static nmea_field_description_t name##_desc []
  DESC(GPGGA) = {{NMF_TIME, NMO_TIME},
	       {NMF_LATITUDE, NMO_DEG}, {NMF_LATITUDE_SIGN, NMO_NORTHSOUTH},
	       {NMF_LONGITUDE, NMO_DEG}, {NMF_LONGITUDE_SIGN, NMO_EASTWEST},
	       {NMF_POSITIONFIX, NMO_DIGIT},
	       {NMF_NUM_SATELLITES, NMO_2DIGIT},
	       {NMF_HDOP, NMO_NUMBER},
	       {NMF_ALTITUDE_MSL, NMO_NUMBER}, {NMF_ALTITUDE_UNIT, NMO_UNIT},
	       {NMF_GEOID_SEPARATION, NMO_NUMBER}, {NMF_GEOID_SEP_UNIT, NMO_UNIT},
	       {NMF_DGPS_AGE, NMO_INT}, {NMF_DGPS_REFID, NMO_INT},
	       {NMF_END, NMO_IGNORE}
  };
  DESC(GPGLL) = {{NMF_LATITUDE, NMO_DEG}, {NMF_LATITUDE_SIGN, NMO_NORTHSOUTH},
	       {NMF_LONGITUDE, NMO_DEG}, {NMF_LONGITUDE_SIGN, NMO_EASTWEST},
	       {NMF_TIME, NMO_TIME},
	       {NMF_GPS_STATUS, NMO_CHAR},
	       {NMF_GPS_MODE, NMO_CHAR},
	       {NMF_END, NMO_IGNORE}
  };
  DESC(GPGSA) = {{NMF_MODE1, NMO_CHAR}, {NMF_MODE2, NMO_DIGIT},
	       {NMF_SATUSED0, NMO_2DIGIT}, {NMF_SATUSED1, NMO_2DIGIT},
	       {NMF_SATUSED2, NMO_2DIGIT}, {NMF_SATUSED3, NMO_2DIGIT},
	       {NMF_SATUSED4, NMO_2DIGIT}, {NMF_SATUSED5, NMO_2DIGIT},
	       {NMF_SATUSED6, NMO_2DIGIT}, {NMF_SATUSED7, NMO_2DIGIT},
	       {NMF_SATUSED8, NMO_2DIGIT}, {NMF_SATUSED9, NMO_2DIGIT},
	       {NMF_SATUSED10, NMO_2DIGIT}, {NMF_SATUSED11, NMO_2DIGIT},
	       {NMF_PDOP, NMO_NUMBER},
	       {NMF_HDOP, NMO_NUMBER}, {NMF_VDOP, NMO_NUMBER},
	       {NMF_END, NMO_IGNORE}
  };
  DESC(GPGSV) = {{NMF_GSV_NUMMSG, NMO_DIGIT}, {NMF_GSV_MSG, NMO_DIGIT},
	       {NMF_NUM_SATELLITES, NMO_2DIGIT},
	       {NMF_SATUSED0, NMO_2DIGIT},
	       {NMF_ELEVATION0, NMO_INT}, {NMF_AZIMUTH0, NMO_INT},
	       {NMF_SNR0, NMO_INT},
	       {NMF_SATUSED1, NMO_2DIGIT},
	       {NMF_ELEVATION1, NMO_INT}, {NMF_AZIMUTH1, NMO_INT},
	       {NMF_SNR1, NMO_INT},
	       {NMF_SATUSED2, NMO_2DIGIT},
	       {NMF_ELEVATION2, NMO_INT}, {NMF_AZIMUTH2, NMO_INT},
	       {NMF_SNR2, NMO_INT},
	       {NMF_SATUSED3, NMO_2DIGIT},
	       {NMF_ELEVATION3, NMO_INT}, {NMF_AZIMUTH3, NMO_INT},
	       {NMF_SNR3, NMO_INT},
	       {NMF_END, NMO_IGNORE}
  };
  DESC(GPRMC) = {{NMF_TIME, NMO_TIME},
	       {NMF_GPS_STATUS, NMO_CHAR},
	       {NMF_LATITUDE, NMO_DEG}, {NMF_LATITUDE_SIGN, NMO_NORTHSOUTH},
	       {NMF_LONGITUDE, NMO_DEG}, {NMF_LONGITUDE_SIGN, NMO_EASTWEST},
	       {NMF_SPEED, NMO_NUMBER}, {NMF_COURSE, NMO_NUMBER},
	       {NMF_DATE, NMO_DATE_YY},
	       {NMF_MAGNETIC_VARIATION, NMO_NUMBER}, {NMF_MAGNETIC_VARIATION_SIGN, NMO_EASTWEST},
	       {NMF_GPS_MODE, NMO_CHAR},
	       {NMF_END, NMO_IGNORE}
  };
  DESC(GPVTG) = {{NMF_COURSE, NMO_NUMBER}, {NMF_COURSE_BASE, NMO_CHAR},
	       {NMF_COURSE2, NMO_NUMBER}, {NMF_COURSE2_BASE, NMO_CHAR},
	       {NMF_SPEED, NMO_NUMBER}, {NMF_SPEED_UNIT, NMO_UNIT},
	       {NMF_SPEED2, NMO_NUMBER}, {NMF_SPEED2_UNIT, NMO_UNIT},
	       {NMF_GPS_MODE, NMO_CHAR},
	       {NMF_END, NMO_IGNORE}
  };
  DESC(GPZDA) = {{NMF_TIME, NMO_TIME},
	       {NMF_DAY, NMO_2DIGIT}, {NMF_MONTH, NMO_2DIGIT},
	       {NMF_YEAR, NMO_4DIGIT},
	       {NMF_UTCOFFSET_H, NMO_2DIGIT}, {NMF_UTCOFFSET_MIN, NMO_2DIGIT},
	       {NMF_END, NMO_IGNORE}
  };
  DESC(PSRF150) = {{NMF_OK_TO_SEND, NMO_DIGIT},
		  {NMF_END, NMO_IGNORE}
  };

typedef struct {
  PGM_P type;
  nmea_field_description_t * fields;
} PROGMEM nmea_description_t;

static nmea_description_t nmea_descriptions [] = {
  {GPGGA_str, GPGGA_desc},
  {GPGLL_str, GPGLL_desc},
  {GPGSA_str, GPGSA_desc},
  {GPGSV_str, GPGSV_desc},
/*{GPMSS_str, GPMSS_desc}, not needed ATM, no DGPS */
  {GPRMC_str, GPRMC_desc},
  {GPVTG_str, GPVTG_desc},
  {GPZDA_str, GPZDA_desc},
  {PSRF150_str, PSRF150_desc},
  {0, 0},
};

#define MAX_FIELD_LEN (16)
static char nmea_field_buf[MAX_FIELD_LEN];
static uint8_t nmea_field_buf_current;

static int8_t nmea_current_field;
static nmea_fieldidx_t nmea_field;
static nmea_field_description_t * nmea_current_description;

nmea_data_t nmea_data;

static int8_t digit2int(char c)
{
  if (c >= '0' && c <= '9') return c-'0';
  return -1;
}

static uint16_t parse_field (nmea_format_t format)
{
  int8_t c0, sign;
  uint16_t result = 0;
  char * p = nmea_field_buf;
  switch (format) {
  case NMO_IGNORE:
    break;
  case NMO_CHAR:
    result = *p;
    break;
  case NMO_4DIGIT:
    if (*p == '\0') {
      break;
    }
    if ((c0 = digit2int (*p++)) < 0) {
      nmea_status = NMS_ERROR;
      break;
    }
    result = c0;
    if (*p == '\0') {
      break;
    }
    if ((c0 = digit2int (*p++)) < 0) {
      nmea_status = NMS_ERROR;
      break;
    }
    result = 10*result+c0;
  case NMO_2DIGIT:
    if (*p == '\0') {
      break;
    }
    if ((c0 = digit2int (*p++)) < 0) {
      nmea_status = NMS_ERROR;
      break;
    }
    result = 10*result+c0;
  case NMO_DIGIT:
    if (*p == '\0') {
      break;
    }
    if ((c0 = digit2int (*p)) < 0) {
      nmea_status = NMS_ERROR;
      break;
    }
    result = 10*result+c0;
    break;
  case NMO_INT:
    if (*p == '-') {
      sign = 1;
      p++;
    } else if (*p == '+') {
      sign = 0;
      p++;
    } else {
      sign = 0;
    }
    while ((c0 = *p++)) {
      c0 = digit2int(c0);
      if (c0 < 0) {
	nmea_status = NMS_ERROR;
	break;
      }
      result = 10*result+c0;
    }
    if (sign) result = -result;
    break;
  case NMO_NORTHSOUTH:
    switch (*p) {
    case 'N': result = 1; break;
    case 'S': result = -1; break;
    case '\0': result = 0; break;
    default: nmea_status = NMS_ERROR; break;
    }
    break;
  case NMO_EASTWEST:
    switch (*p) {
    case 'E': result = 1; break;
    case 'W': result = -1; break;
    case '\0': result = 0; break;
    default: nmea_status = NMS_ERROR; break;
    }
    break;
  case NMO_UNIT:
    result = *p;
    break;

  case NMO_NUMBER:
  case NMO_TIME:
  case NMO_DATE_YY:
  case NMO_DEG:
    nmea_status = NMS_ERROR;
    break;
  }
  return result;
}

static void parse_field_dword (nmea_format_t format, uint32_t * pfield)
{
  char * p = nmea_field_buf;
  int8_t c0 = 0, c1 = 0;
  uint32_t result = 0;
  uint16_t result1 = 0;
  uint32_t frac_result = 0;
  uint32_t frac_divider = 1;

  switch (format) {
  case NMO_IGNORE:
    break;
  case NMO_CHAR:
  case NMO_DIGIT: case NMO_2DIGIT: case NMO_4DIGIT:
  case NMO_INT:
  case NMO_NORTHSOUTH: case NMO_EASTWEST:
  case NMO_UNIT:
    nmea_status = NMS_ERROR;
    break;

  case NMO_NUMBER:
    while ((c0 = *p++)) {
      if (c0 == '.') break;
      result1 = result1 * 10 + digit2int (c0);
    }
    while ((c0 = *p++)) {
      if ((frac_result * 10)>>16 >= 1) break;
      frac_result = frac_result * 10 + digit2int (c0);
      frac_divider *= 10;
    }
    result = (result << 16) | (frac_result<<16 / frac_divider);
    break;
  case NMO_DEG:
  case NMO_TIME:
    /* scan for '.' */
    c0 = -1;
    while (*p) {
      c1++;
      switch (*p) {
      case '.':
	c0 = p - nmea_field_buf;
	break;
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	break;
      default:
	nmea_status = NMS_ERROR;
	break;
      }
      p++;
    }
    if (nmea_status == NMS_ERROR) break;
    p = nmea_field_buf;
    if (c0 == -1) c0 = c1;
    /* now p[c0] == '.' || p[c0] == '\0' */
    switch (c0) {
      /* fall-through is deliberate */
    default:
    case 6:  /* 10hours*/
      result += 36000*digit2int(p[c0-6]);
    case 5:  /* hours | 100degrees */
      result += (format == NMO_TIME ? 3600 : 6000)*digit2int(p[c0-5]);
    case 4:  /* 10minutes | 10degrees */
      result += 600*digit2int(p[c0-4]);
    case 3:  /* minutes | degrees*/
      result += 60*digit2int(p[c0-3]);
    case 2:  /* 10seconds | 10minutes*/
      result += 10*digit2int(p[c0-2]);
    case 1:  /* seconds */
      result += digit2int (p[c0-1]);
    case 0:
      break;
    }
    c0++; /* point to char after '.' */
    while (c0 < c1 && frac_divider < (1<<(format == NMO_TIME ? 15 : 16))) {
      frac_result *= 10;
      frac_result += digit2int (p[c0++]);
      frac_divider *= 10;
    }
    /* compose time value */
    if (format == NMO_TIME) {
      union {
	gps_time_t time;
	uint32_t result;
      } time_result;
      time_result.time.seconds = result;
      time_result.time.frac_seconds = ((frac_result<<15) / frac_divider);
      result = time_result.result;
    } else { /* format == NMO_DEG */
      result = (result << 16)
	| (((uint32_t)1<<16) * frac_result / frac_divider);
    }
    break;
  case NMO_DATE_YY:
    {
      c0 = strlen (p);
      union {
	gps_date_t date;
	uint32_t result;
      } date_result;
      date_result.result = *pfield;
      date_result.date.year = (date_result.date.year/100)*100;
      date_result.date.month = 0;
      date_result.date.day = 0;
      switch (c0) {
	/* fall-through is deliberate */
      default:
      case 6:
	date_result.date.year += digit2int (p[c0-6]) * 10;
      case 5:
	date_result.date.year += digit2int (p[c0-5]);
      case 4:
	date_result.date.month += digit2int (p[c0-4]) * 10;
      case 3:
	date_result.date.month += digit2int (p[c0-3]);
      case 2:
	date_result.date.day += digit2int (p[c0-2]) * 10;
      case 1:
	date_result.date.day += digit2int (p[c0-1]);
      case 0:
	break;
      }
      result = date_result.result;
    }
    break;
  }
  *pfield = result;
}

static void nmea_handle_field (nmea_field_t field, nmea_format_t format)
{
  int32_t tmp;
  uint16_t sign;
  uint8_t satnum;

  switch (field) {
  case NMF_END:
    nmea_status = NMS_ERROR;
    return;
  case NMF_IGNORE:
    break;
  case NMF_TIME:
    parse_field_dword (format, (uint32_t*)&nmea_data.time);
    nmea_data.status.time = 1;
    break;
  case NMF_DATE:
    parse_field_dword (format, (uint32_t*)&nmea_data.date);
    nmea_data.status.year2 = 1;
    nmea_data.status.month = 1;
    nmea_data.status.day = 1;
    break;
  case NMF_DAY:
    nmea_data.date.day = parse_field (format);
    nmea_data.status.day = 1;
    break;
  case NMF_MONTH:
    nmea_data.date.month = parse_field (format);
    nmea_data.status.month = 1;
    break;
  case NMF_YEAR:
    nmea_data.date.year = parse_field (format);
    nmea_data.status.year2 = 1;
    nmea_data.status.year4 = 1;
    break;
  case NMF_UTCOFFSET_H:
    nmea_data.utc_offset.hours = parse_field (format);
    nmea_data.status.utc_offset_hours = 1;
    break;
  case NMF_UTCOFFSET_MIN:
    nmea_data.utc_offset.minutes = parse_field (format);
    nmea_data.status.utc_offset_minutes = 1;
    break;
  case NMF_LATITUDE:
    parse_field_dword (format, &nmea_data.latitude);
    nmea_data.status.latitude = 1;
    break;
  case NMF_LATITUDE_SIGN:
    sign = parse_field (format);
    if (sign) {
      nmea_data.latitude_sign = (sign < 0);
      nmea_data.status.latitude_sign = 1;
    }
    break;
  case NMF_LONGITUDE:
    parse_field_dword (format, &nmea_data.longitude);
    nmea_data.status.longitude = 1;
    break;
  case NMF_LONGITUDE_SIGN:
    sign = parse_field (format);
    if (sign) {
      nmea_data.longitude_sign = (sign < 0);
      nmea_data.status.longitude_sign = 1;
    }
    break;
  case NMF_ALTITUDE_MSL:
    parse_field_dword (format, &nmea_data.altitude);
    nmea_data.status.altitude = 1;
    break;
  case NMF_GEOID_SEPARATION:
    parse_field_dword (format, &nmea_data.geoid_separation);
    nmea_data.status.geoid_separation = 1;
    break;
  case NMF_ALTITUDE_UNIT:
    switch (parse_field (format)) {
    case 'M':
      nmea_data.altitude_unit = UNIT_METER;
      nmea_data.status.altitude_unit = 1;
      break;
    case 'F': /* maybe */
      nmea_data.altitude_unit = UNIT_FOOT;
      nmea_data.status.altitude_unit = 1;
      break;
    }
    break;
  case NMF_GEOID_SEP_UNIT:
    switch (parse_field (format)) {
    case 'M':
      nmea_data.geoid_separation_unit = UNIT_METER;
      nmea_data.status.geoid_separation_unit = 1;
      break;
    case 'F': /* maybe */
      nmea_data.geoid_separation_unit = UNIT_FOOT;
      nmea_data.status.geoid_separation_unit = 1;
      break;
    }
    break;
  case NMF_DGPS_AGE:
    nmea_data.dgps_age = parse_field (format);
    nmea_data.status.dgps_age = 1;
    break;
  case NMF_DGPS_REFID:
    nmea_data.dgps_refid = parse_field (format);
    nmea_data.status.dgps_refid = 1;
    break;
  case NMF_POSITIONFIX:
    nmea_data.positionfix = parse_field (format);
    nmea_data.status.positionfix = 1;
    break;
  case NMF_NUM_SATELLITES:
    nmea_data.numsats = parse_field (format);
    nmea_data.status.numsats = 1;
    break;
  case NMF_PDOP:
    parse_field_dword (format, &nmea_data.pdop);
    nmea_data.status.pdop = 1;
    break;
  case NMF_HDOP:
    parse_field_dword (format, &nmea_data.hdop);
    nmea_data.status.hdop = nmea_field_buf[0] != '\0';
    break;
  case NMF_VDOP:
    parse_field_dword (format, &nmea_data.vdop);
    nmea_data.status.vdop = 1;
    break;
  case NMF_OK_TO_SEND:
    nmea_data.ok_to_send = parse_field (format);
    nmea_data.status.ok_to_send = 1;
    break;
  case NMF_GSV_NUMMSG:
    nmea_data.status.gsv_num_msg = parse_field (format);
    break;
  case NMF_GSV_MSG:
    nmea_data.status.gsv_msg_num = parse_field (format);
    break;
  case NMF_SATUSED0:
  case NMF_SATUSED1:
  case NMF_SATUSED2:
  case NMF_SATUSED3:
    if (nmea_data.status.gsv_num_msg) {
      /* this is a GSV sentence */
      switch (field) {
      case NMF_SATUSED0:
	satnum = 0;
	break;
      case NMF_SATUSED1:
	satnum = 1;
	break;
      case NMF_SATUSED2:
	satnum = 2;
	break;
      case NMF_SATUSED3:
	satnum = 3;
	break;
      default: /* shouldn't happen */
	break;
      }
      satnum += 4*(nmea_data.status.gsv_msg_num-1);
      nmea_data.status.gsv_satnum = satnum; /* for following fields */
      nmea_data.satposition[satnum].svid = parse_field (format);
      nmea_data.status.svid |= (1<<satnum);
      break;
    }
  case NMF_SATUSED4:
  case NMF_SATUSED5:
  case NMF_SATUSED6:
  case NMF_SATUSED7:
  case NMF_SATUSED8:
  case NMF_SATUSED9:
  case NMF_SATUSED10:
  case NMF_SATUSED11:

    switch (field) {
    case NMF_SATUSED0:
      satnum = 0;
      break;
    case NMF_SATUSED1:
      satnum = 1;
      break;
    case NMF_SATUSED2:
      satnum = 2;
      break;
    case NMF_SATUSED3:
      satnum = 3;
      break;
    case NMF_SATUSED4:
      satnum = 4;
      break;
    case NMF_SATUSED5:
      satnum = 5;
      break;
    case NMF_SATUSED6:
      satnum = 6;
      break;
    case NMF_SATUSED7:
      satnum = 7;
      break;
    case NMF_SATUSED8:
      satnum = 8;
      break;
    case NMF_SATUSED9:
      satnum = 9;
      break;
    case NMF_SATUSED10:
      satnum = 10;
      break;
    case NMF_SATUSED11:
      satnum = 11;
      break;
    default: /* shouldn't happen */
      break;
    }
    nmea_data.satposition[satnum].svid = parse_field (format);
    nmea_data.status.svid |= (1<<satnum);
    break;
  case NMF_ELEVATION0:
  case NMF_ELEVATION1:
  case NMF_ELEVATION2:
  case NMF_ELEVATION3:
    satnum = nmea_data.status.gsv_satnum;
    nmea_data.satposition[satnum].ele = parse_field (format);
    nmea_data.status.satpos_ele |= (1<<satnum);
    break;
  case NMF_AZIMUTH0:
  case NMF_AZIMUTH1:
  case NMF_AZIMUTH2:
  case NMF_AZIMUTH3:
    satnum = nmea_data.status.gsv_satnum;
    nmea_data.satposition[satnum].azi = parse_field (format);
    nmea_data.status.satpos_azi |= (1<<satnum);
    break;
  case NMF_SNR0:
  case NMF_SNR1:
  case NMF_SNR2:
  case NMF_SNR3:
    satnum = nmea_data.status.gsv_satnum;
    nmea_data.satposition[satnum].snr = parse_field (format);
    nmea_data.status.satpos_snr |= (1<<satnum);
    break;
  case NMF_MODE1:
    nmea_data.mode1 = parse_field (format);
    nmea_data.status.mode1 = 1;
    break;
  case NMF_MODE2:
    nmea_data.mode2 = parse_field (format);
    nmea_data.status.mode2 = 1;
    break;
  }
  return;
}

static void nmea_finish_field (void)
{
  nmea_field_buf[nmea_field_buf_current] = '\0';
  if (nmea_current_field == -1) {
    /* type of message */
    nmea_description_t * d;
    PGM_P msg;
    for (d = nmea_descriptions; (msg = (PGM_P)pgm_read_word(&d->type)); d++) {
      if (!strcmp_P (nmea_field_buf, msg)) {
	/* found match */
	nmea_current_description = (nmea_field_description_t *)pgm_read_word(&d->fields);
	nmea_field = NMP_FIELDS;
	nmea_current_field = 0;
	if (nmea_start_callback)
	  (*nmea_start_callback)(nmea_field_buf);
	return;
      }
    }
    nmea_status = NMS_ERROR;
    return;
  } else {
    /* message field */
    nmea_field_description_t * d = nmea_current_description+nmea_current_field;
    nmea_handle_field (pgm_read_field (d), pgm_read_format (d));
    if (pgm_read_field(d) != NMF_END) nmea_current_field++;
    return;
  }
}

static void nmea_status_start (uint8_t b)
{
  nmea_status = NMS_STARTED;
  nmea_data.csum = 0;
  nmea_data.time.seconds = 0; nmea_data.time.frac_seconds = 0;
  nmea_data.latitude = 0;
  nmea_data.longitude = 0;
  nmea_data.quality = 0;
  nmea_data.numsats = 0;
  nmea_data.status.valid = 0;
  nmea_data.status.error = 0;
  memset (&nmea_data.status, 0, sizeof (nmea_data.status));
  nmea_field_buf_current = 0;
  nmea_current_field = -1;
  memset (nmea_field_buf, 0, sizeof (nmea_field_buf));
}

static void nmea_csum (uint8_t b)
{
  nmea_data.csum ^= b;
}

void nmea_serial_in_handler (uint8_t b)
{
  /* checksum handling */
  switch (nmea_status) {
  case NMS_NEW:
  case NMS_ERROR:
  case NMS_CR_SEEN:
  case NMS_LF_SEEN:
  case NMS_CSUM1:
  case NMS_CSUM2:
  case NMS_CSUM_SEEN:
    break;
  case NMS_STARTED:
  default:
    if (b != '*' /*&& b != NMEA_MSG_START*/)
      nmea_csum (b);
    break;
  }

  switch (nmea_status) {
  case NMS_NEW:
  case NMS_ERROR:
    switch (b) {
    case '\r':
      nmea_status = NMS_CR_SEEN;
      break;
    case '\n':
      nmea_status = NMS_LF_SEEN;
      break;
    case NMEA_MSG_START:
      nmea_status_start (b);
      break;
    default: /* ignore any other */
      break;
    }
    break; /* NMS_NEW */
  case NMS_CR_SEEN:
    switch (b) {
    case '\n':
      nmea_status = NMS_LF_SEEN;
      break;
    default:
      nmea_status = NMS_ERROR;
      break;
    }
    break; /* NMS_CR_SEEN */
  case NMS_LF_SEEN:
    switch (b) {
    case NMEA_MSG_START:
      nmea_status_start (b);
      break;
    default:
      nmea_status = NMS_ERROR;
    }
    break; /* NMS_LF_SEEN */
  case NMS_STARTED:
    switch (b) {
    case ',':
      nmea_finish_field ();
      memset (nmea_field_buf, 0, sizeof (nmea_field_buf));
      nmea_field_buf_current = 0;
      break;
    case '*':
      nmea_finish_field ();
      memset (nmea_field_buf, 0, sizeof (nmea_field_buf));
      nmea_field_buf_current = 0;
      nmea_status = NMS_CSUM1;
      break;
    case '\r':
      nmea_status = NMS_CR_SEEN;
      break;
    case '\n':
      nmea_status = NMS_LF_SEEN;
      break;
    default:
      if (nmea_field_buf_current < sizeof(nmea_field_buf)-1) {
	nmea_field_buf[nmea_field_buf_current++] = b;
      } else {
	nmea_status = NMS_ERROR;
      }
      break;
    }
    break; /* NMS_STARTED */
  case NMS_CSUM1:
    if (b >= '0' && b <= '9')
      nmea_data.csum_read = (b - '0') << 4;
    else if (b >= 'A' && b <= 'F')
      nmea_data.csum_read = (b - 'A' + 10) << 4;
    else if (b >= 'a' && b <= 'f')
      nmea_data.csum_read = (b - 'a' + 10) << 4;
    else {
      nmea_status = NMS_ERROR;
      break;
    }
    nmea_status = NMS_CSUM2;
    break; /* NMS_CSUM1 */
  case NMS_CSUM2:
    if (b >= '0' && b <= '9')
      nmea_data.csum_read |= (b - '0');
    else if (b >= 'A' && b <= 'F')
      nmea_data.csum_read |= (b - 'A' + 10);
    else if (b >= 'a' && b <= 'f')
      nmea_data.csum_read |= (b - 'a' + 10);
    else {
      nmea_status = NMS_ERROR;
      break;
    }
    nmea_status = NMS_CSUM_SEEN;
    break; /* NMS_CSUM2 */
  case NMS_CSUM_SEEN:
    switch (b) {
    case '\r':
      nmea_status = NMS_CR_SEEN;
      nmea_data.status.valid = !nmea_data.status.error
	&& nmea_data.csum == nmea_data.csum_read;
      if (nmea_data.status.valid) {
	put_nmea_event (&nmea_data);
      }
      break;
    default:
      nmea_status = NMS_ERROR;
      nmea_data.status.error = 1;
      break;
    }
    break; /* NMS_CSUM_SEEN */
  }
}
